#!/usr/local/bin/python2.6
#encoding: utf-8

##
## Este programa pega um array com intensidades de pixel de Ncor
## canais, em linhas com Npix colunas, e obtidos de Npos posições
## diferentes, encontra picos em cada linha e grava em um arquivo de
## saída um alista contendo as eviências em cada posição, isto é, uma
## lista com os valores do inverso da coluna em que se localizam os
## picos detectados.


import numpy
numpy.set_printoptions(precision=3)


import sys, time
import pp

import itertools





##
## Calculates vector of correspondence estimates given evidences and
## model parameters
def correspondence_likelyhood(Wa, pa, Wb, pb, vj, xj, sigma=0.3):  ##sei lá o valor bom pra esse sigma...
    coef = -1/(2*sigma**2)

    pred_a = vj*(xj-pa)
    pred_b = vj*(xj-pb)

    out=[]

    ptlist= itertools.product(Wa, Wb)

    for wa,wb in ptlist:

        out.append(0.0 if (wa>0) ^ (wb<wa) or ((wa>0) != (wb>0))
                   else \
                       numpy.exp(coef * (
                                           (wa-pred_a)**2.0
                                         + (wb-pred_b)**2.0
                                        ))
                   )
    return numpy.array(out)






## Updates map parameters from a set of evidences
## Includes calculation of the expectation of corerspondence
def get_new_theta(Z, xj, vj, sigma=0.3):
    Npos = len(Z)

    ## start function
    new_vj_num=0
    new_vj_den=0
    new_xj_num=0
    ## new_xj_den, not, because it is precisely the vj_num

    for pa,pb in itertools.combinations( range(Npos) , 2  ):
        deltap = pb-pa
        ## Seleciona só pares a distâncias dentro desta faixa
        p_max=1e20
        p_min=8
        if deltap>p_max or deltap<p_min:
            continue

        ## Calculate the corresponcence factor

        Wa = Z[pa]
        Wb = Z[pb]
        c = correspondence_likelyhood(Wa, pa, Wb, pb, vj, xj, sigma)

        wa_minus_wb = numpy.array([wa - wb for wa, wb in itertools.product( Wa, Wb )])
        pbwa_minus_pawb = numpy.array([pb*wa - pa*wb for wa, wb in itertools.product( Wa, Wb )])

        new_vj_num += deltap    * numpy.sum(c * (wa_minus_wb) )
        new_vj_den += deltap**2 * numpy.sum(c)
        new_xj_num += deltap    * numpy.sum(c * (pbwa_minus_pawb) )

        do_debug = False
        if do_debug:
            print '(xj, vj)=', (xj, vj)
            print 'pb, pa', pb,pa
            print 'Wa',Wa
            print 'Wb',Wb
            print c
            print 'wa-wb',wa_minus_wb
            print 'sum c*(wa-wb)', numpy.sum(c*wa_minus_wb)
            print 'sum c',numpy.sum(c)
            print 'pbwa-pawb', pbwa_minus_pawb
            print 'sum c*(pbwa-pawb)', numpy.sum(c*pbwa_minus_pawb)
            print 'newx',new_xj_num/new_vj_num
            print 'newy',new_vj_num/new_vj_den
            print '*'*80
            print 
            print 
            print 


    ## Returns new xj=xj_num/vj_num and new vj=vj_num/vj_den
    return numpy.array([ new_xj_num/new_vj_num , new_vj_num/new_vj_den])







##
## Return all possible map parameter values supposing all viable matches
def get_possible_params(Wa, pa, Wb, pb):
    out=[]

    ptlist= itertools.product(Wa, Wb)

    for wa,wb in ptlist:
        if (wb>0) ^ (wb<wa) or wa==wb:
            continue
            
        out.append( [(wa*pb-wb*pa)/ (wa-wb), (wa-wb)/(pb-pa) ] )
                   
    return numpy.array(out)

## Generates statistics about possible mark locations
def map_statistics(Z):
    Npos = len(Z)

    ## start function
    new_vj_num=0
    new_vj_den=0
    new_xj_num=0
    ## new_xj_den, not, because it is precisely the vj_num


    cc = numpy.zeros((2,2))
    m =  numpy.zeros((1,2))
    N=0.0
    for pa,pb in itertools.combinations( range(Npos) , 2  ):
        c=get_possible_params(Z[pa], pa, Z[pb], pb) 
        m+=numpy.sum(c,0)
        cc += numpy.dot(c.T,c)
        N  += c.shape[0]

    

    return [m/N, cc/N]
    




## Prints the values of the weights relative to the probability of correspondence
def print_corresp(Z, Theta):

    Npos = len(Z)
    for theta in Theta:
        (xj, vj)=theta

        for pp in itertools.combinations( range(Npos) , 2  ):
            pa, pb=pp
            deltap = pb-pa
            ## Seleciona só pares a distâncias dentro desta faixa
            p_max=1e30
            p_min=1
            if deltap>p_max or p_min<p_min:
                continue

            print xj, vj, pa, pb

        ## Calculate the corresponcence factor

            Wa = Z[pa]
            Wb = Z[pb]
            deltap = pb-pa
            c = correspondence_likelyhood(Wa, pa, Wb, pb, vj, xj)

            WW = [z for z in itertools.product( range(len(Wa)), range(len(Wb)) )]
            #onde = numpy.where(c > 1e-60)[0]
            onde = numpy.where(c>0)[0]
        
            for o in onde:
                print WW[o], "% 6.2f % 6.2f   %.2G"%(Wa[WW[o][0]], Wb[WW[o][1]], c[o],)









##
##
## Carrega a lista de evidências, Z, de um arquivo
import pickle
infile = open(sys.argv[1], 'rb')
Z=pickle.load(infile)
infile.close()



## 
## Inicializa o mapa

#Theta = numpy.array([ 
#[1613.0, -0.00186335403727],
#[144, 0.95],
#         ], dtype='float32')

#Theta = numpy.array( [z for z in itertools.product(
#        numpy.arange(10, 40, 8.0)**2,
#        numpy.arange(-0.1, 1.0, 0.05)
#        )]
#                     , dtype='float32')

Theta= numpy.array([

        [  9.61123866e+01,   1.34051779e-01],
        [  9.61192989e+01,   3.37307510e-01],
        [  9.44523380e+01,   4.41956418e-01],

        [  5.58699088e+02,  -5.42623912e-03],
        [  5.58699088e+02,  -6.42623912e-03],
        [  4.58699088e+02,  -5.42623912e-03],
        [  4.58699088e+02,  -6.42623912e-03],


        [  1.04174871e+02,   2.98760213e-01],
        [  1.40234504e+02,   1.21763159e-01],
        [  1.50923797e+02,   2.23421080e-01],
        [  1.89341946e+02,   9.79620980e-02],
        [  2.09781953e+02,   3.23747805e-02],
])



print Theta




Nmrc, lixo = Theta.shape

Theta_label = numpy.array(range(Nmrc))





##
##
## Parallel Python stuff
# tuple of all parallel python servers to connect with
ppservers = ()
ppservers = ("143.107.165.142",)
#ppservers = ("201.58.54.81:2002",)
#ppservers = ("akira.ath.cx:2002","akira.ath.cx:3000")

#ppservers = ("143.107.165.142",)
#ppservers = ("127.0.0.1",)

if len(sys.argv) > 1:
    ncpus = 4
    # Creates jobserver with ncpus workers
    job_server = pp.Server(ncpus, ppservers=ppservers, secret='424242',proto=2,loglevel=30)
else:
    # Creates jobserver with automatically detected number of workers
    job_server = pp.Server(ppservers=ppservers,secret='424242',proto=2,loglevel=30)

print "Starting pp with", job_server.get_ncpus(), "workers"





















##
##
## Levanta estatística de possíveis marcos
calc_estat=True

if calc_estat:
    m,cc=map_statistics(Z)
    ccc = cc-numpy.dot(m.T,m)
    print '-'*70
    print "Estatística de valores possíveis"
    print 'µx', m[0,0], 'µv', m[0,1],  'µy', 1/m[0,1]
    print cc
    print ccc
    print 'σx', numpy.sqrt(ccc[0,0]), 'σv', numpy.sqrt(ccc[1,1])
    print '-'*70
    print















###
### E-step
###

## O cálculo deveria ser feito tomando-se a função de verossimilhança
## de c_ij dados, as medições e os parâmetros do modelo, e então
## normalizá-la. A normalização porém não influencia no funcionamenbto
## do algoritmo. Portanto vamos apenas calcular um vetor com valores
## proporcionais.

## ...E no final das contas basta fazer:
## E[c_ij] = exp( -(w_i-v_j(x_j-p_i))**2/2/sig**2  )

###
### M-step (actually "EM-step", but the separate E is unncessary)
###




print "Theta:"
for t in Theta:
    print t

for Niter in range(1000):
    
    sigma = 1.0 # 1.0/float(Niter+3)

    ## Lista de jobs
    jobs = [ job_server.submit(get_new_theta,
                                    (Z, Theta[j,0], Theta[j,1], sigma),
                                    (correspondence_likelyhood,),
                                    ("numpy","itertools"))
             for j in range(Nmrc)]

    start_time = time.time()
    newTheta = numpy.array([ job() for job in jobs ])
    elaps = time.time() - start_time

    theta_ok = numpy.array([numpy.isfinite(k).all() for k in newTheta])
    
    ## Elimina marcos se já existir um que é só "1%" da distância.
    ## O certo é levantar um limiar de erro pra cada dimensão...
    for j,k in itertools.combinations( range(Nmrc) , 2  ):
        if not (theta_ok[j] and theta_ok[k]):
            continue
        rel = numpy.abs(1-newTheta[j] / newTheta[k])
        if (rel<0.01).all():
            theta_ok[k] = False
            print " ** Eliminando marco %d, (%d repetido) "%(Theta_label[k],Theta_label[j]), newTheta[k]

    if not theta_ok.any():
        print " ** Matamos todos os marcos! :("
        break

    Theta = numpy.array( newTheta[ theta_ok,: ] )
    Theta_label = Theta_label[ theta_ok,: ] 

    Nmrc, lixo = Theta.shape
        

    print "ITERATION %d  --- Duration: %fs"%( Niter, elaps)
    print 'σ=',sigma
    print "Theta [%d]:"%Nmrc
    for t in max(xrange(Nmrc),6):
        print "% 3d"%Theta_label[t], Theta[t]**[1,1]

    job_server.print_stats()
    print 70*'='
    print

    ##
    ##
    ## Salva a saída 
    output = open(sys.argv[2], 'wb')
    pickle.dump((Niter, Theta) , output, -1)
    output.close()



#for z in Z[:20]:
#    print z
#print_corresp(Z, Theta)


##
##
## Salva a saída 
output = open(sys.argv[2], 'wb')
pickle.dump((Niter, Theta) , output, -1)
output.close()











 #  0 [  5.587e+02  -5.426e-03]
 #  2 [  2.098e+02   3.237e-02]
 #  4 [ 96.112   0.134]
 #  6 [ 150.924    0.223]
 #  7 [  1.402e+02   1.218e-01]
 #  8 [ 96.119   0.337]
 #  9 [ 94.452   0.442]
 # 10 [ 104.175    0.299]
 # 25 [  1.893e+02   9.796e-02]
